/**
 * \file taggingscheme.cpp
 * Class for reading, writing and storing tag manipulation scheme.
 *
 * \b Project: Qoobar
 * \author Alex Novichkov
 * \date 17 Jan 2012
 *
 * Copyright (C) 2012  Alex Novichkov
 *
 * This file is part of Qoobar.
 *
 * Qoobar is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Qoobar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "taggingscheme.h"

#include <QFile>
#include <QXmlStreamReader>
#include <QXmlStreamWriter>
#include "qoobarglobals.h"
#include <QFileInfo>
#include "applicationpaths.h"

#include "enums.h"

TaggingScheme::TaggingScheme(const QString &filePath)
    : m_filePath(filePath)
{DD;
    m_fieldsForParsing.resize(TAGSCOUNT);
    m_fields.resize(TAGSCOUNT);
}

void TaggingScheme::setFilePath(const QString &filePath)
{DD;
    m_filePath=filePath;
}

void TaggingScheme::read(bool nameOnly)
{DD;
    QFile f(m_filePath);
    if (!f.exists()) {
        f.setFileName(QSL(":/src/default.xml"));
    }
    if (!f.open(QFile::ReadOnly | QFile::Text))
        return;

    QXmlStreamReader xml(&f);
    int id=0;
    TagType type=UNKNOWN;

    while (!xml.atEnd()) {
        xml.readNext();
        if (xml.isStartElement()) {
            QStringRef name=xml.name();
            if (name == "scheme") {
                setName(xml.attributes().value(QSL("name")).toString());
                if (nameOnly) return;
            }
            else if (name == "tag")
                id = xml.attributes().value(QSL("id")).toString().toInt();
            else if (name == "format")
                type = (TagType)xml.attributes().value(QSL("id")).toString().toInt();
            else if (name=="field") {
                if (type!=UNKNOWN) {
                    QString s=xml.readElementText();
                    m_fieldsForParsing[type].insert(s,id);
                    m_fields[type].insert(id, s);
                }
            }
        }
    }
}

bool TaggingScheme::save()
{DD;
    //try to save into user settings folder
    m_filePath = QString("%1/%2").arg(ApplicationPaths::userSchemesPath())
            .arg(QFileInfo(m_filePath).fileName());

    QFile f(m_filePath);
    if (f.open(QFile::WriteOnly)) {
        QXmlStreamWriter stream(&f);
        stream.setAutoFormatting(true);
        stream.writeStartDocument(QSL("1.0"));

        stream.writeStartElement(QSL("scheme"));
        stream.writeAttribute(QSL("name"), m_name);
        for (int i=0; i<TAGSCOUNT; ++i) {
            stream.writeStartElement(QSL("tag"));
            stream.writeAttribute(QSL("name"),Qoobar::untranslatedTagByID(i));
            stream.writeAttribute(QSL("id"),QString::number(i));
            for (int j=0; j<5; ++j) {
                stream.writeStartElement(QSL("format"));
                stream.writeAttribute(QSL("id"),QString::number(j));
                QStringList fieldsList = fields(i,(TagType)j);
                Q_FOREACH (const QString &field, fieldsList) {
                    stream.writeTextElement(QSL("field"),field);
                }
                stream.writeEndElement(); //"format"
            }
            stream.writeEndElement(); //"tag"
        }
        stream.writeEndDocument();
        return true;
    }
    return false;
}

int TaggingScheme::search(TagType tagType, const QString &field)
{DD;
    return m_fieldsForParsing.at(tagType).value(field,-1);
}

bool TaggingScheme::equal(TagType tagType,const QString &field, int id) const
{DD;
    QMultiHash<QString, int> hash=m_fieldsForParsing.at(tagType);
    return (hash.contains(field, id) && hash.contains(field, id+1));

    //QVector<int> fieldsIDs = m_fieldsForParsing.value(tagType).values(field);
    ///return (fieldsIDs.contains(id) && fieldsIDs.contains(id+1));
}

QStringList TaggingScheme::fields(int tagID, TagType tagType) const
{DD;
    return m_fields.at(tagType).values(tagID);
}
QVector<int> TaggingScheme::tagIDs(const QString &field, TaggingScheme::TagType tagType) const
{DD;
    return m_fieldsForParsing.at(tagType).values(field).toVector();
}
void TaggingScheme::clear()
{DD;
    m_fieldsForParsing.clear();
    m_fields.clear();
    m_fieldsForParsing.resize(TAGSCOUNT);
    m_fields.resize(TAGSCOUNT);
}

void TaggingScheme::setFields(int tagID, TagType tagType, const QStringList &fields)
{DD;
    Q_FOREACH (const QString &field, fields) {
        m_fieldsForParsing[tagType].insert(field,tagID);
        m_fields[tagType].insert(tagID,field);
    }
}

QString TaggingScheme::name() const
{DD;
    return m_name;
}

void TaggingScheme::setName(const QString &name)
{DD;
    m_name=name;
}
